@@ -8,7 +8,7 @@ import threading
 import uuid
 import warnings
 from email.mime.base import MIMEBase
-from typing import TYPE_CHECKING, Dict, Iterable, Optional, Tuple, Union
+from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Tuple, Union
 
 from django.conf import settings
 from django.core.exceptions import ImproperlyConfigured
@@ -34,7 +34,12 @@ from sendgrid.helpers.mail import (
 )
 
 from sendgrid_backend.signals import sendgrid_email_sent
-from sendgrid_backend.util import SENDGRID_5, SENDGRID_6, get_django_setting
+from sendgrid_backend.util import (
+    SENDGRID_5,
+    SENDGRID_6,
+    dict_to_personalization,
+    get_django_setting,
+)
 
 DjangoAttachment = Union[Tuple[str, Union[bytes, str], str], MIMEBase]
 
@@ -240,6 +245,101 @@ class SendgridBackend(BaseEmailBackend):
             name = None
         return addr, name
 
+    def _build_sg_personalization(
+        self,
+        msg: EmailMessage,
+        extra_headers: Iterable[Header],
+        to: Optional[List[str]] = None,
+        existing_personalizations: Optional[Personalization] = None,
+    ) -> Personalization:
+        """
+        Constructs a Sendgrid Personalization instance / row for the given recipients.
+
+        If no "per row" personalizations are provided, the personalization data is populated
+        from msg.
+
+        Args:
+            msg: The base Django Email message object - used to fill (missing) personalization data
+            extra_headers: The non "reply-to" headers for the personalization.
+            to: The email addresses for the given personalization.
+            existing_personalizations: Personalization data, eg. dynamic_template_data or substitutions.
+                A given value should have key equivalent to the corresponding EmailMessage attr
+
+
+        Returns:
+            A sendgrid personalization instance
+        """
+
+        if existing_personalizations is None:
+            if not to:
+                raise ValueError(
+                    "Either msg.to or msg.personalizations (with recipients) must be set"
+                )
+
+            personalization = Personalization()
+            for addr in to:
+                personalization.add_to(Email(*self._parse_email_address(addr)))
+
+        elif existing_personalizations.tos:
+            personalization = existing_personalizations
+        else:
+            raise ValueError("Each msg personalization must have recipients")
+
+        if not personalization.ccs:
+            for addr in msg.cc:
+                personalization.add_cc(Email(*self._parse_email_address(addr)))
+
+        if not personalization.bccs:
+            for addr in msg.bcc:
+                personalization.add_bcc(Email(*self._parse_email_address(addr)))
+
+        if not personalization.custom_args:
+            for k, v in getattr(msg, "custom_args", {}).items():
+                personalization.add_custom_arg(CustomArg(k, v))
+
+        if self._is_transaction_template(msg):
+            if personalization.subject or msg.subject:
+                logger.warning(
+                    "Message subject is ignored in transactional template, "
+                    "please add it as template variable (e.g. {{ subject }}"
+                )
+                # See https://github.com/sendgrid/sendgrid-nodejs/issues/843
+
+        if not personalization.subject:
+            personalization.subject = msg.subject
+
+        for header in extra_headers:
+            personalization.add_header(header)
+
+        # write through the send_at attribute
+        if hasattr(msg, "send_at"):
+            if not isinstance(msg.send_at, int):
+                raise ValueError(
+                    "send_at must be an integer, got: {}; "
+                    "see https://sendgrid.com/docs/API_Reference/SMTP_API/scheduling_parameters.html#-Send-At".format(
+                        type(msg.send_at)
+                    )
+                )
+            personalization.send_at = msg.send_at
+
+        if hasattr(msg, "template_id"):
+            if not personalization.substitutions:
+                for k, v in getattr(msg, "substitutions", {}).items():
+                    personalization.add_substitution(Substitution(k, v))
+
+            dtd = personalization.dynamic_template_data or getattr(
+                msg, "dynamic_template_data", None
+            )
+            if dtd:
+                if SENDGRID_5:
+                    logger.warning(
+                        "dynamic_template_data not available in sendgrid version < 6"
+                    )
+                else:
+                    personalization.dynamic_template_data = dtd
+
+        return personalization
+
     def _build_sg_mail(self, msg: EmailMessage) -> Dict:
         """
         Serializes a Django EmailMessage into its JSON representation.
@@ -250,35 +350,12 @@ class SendgridBackend(BaseEmailBackend):
 
         mail.from_email = Email(*self._parse_email_address(msg.from_email))
 
-        personalization = Personalization()
-        for addr in msg.to:
-            personalization.add_to(Email(*self._parse_email_address(addr)))
-
-        for addr in msg.cc:
-            personalization.add_cc(Email(*self._parse_email_address(addr)))
-
-        for addr in msg.bcc:
-            personalization.add_bcc(Email(*self._parse_email_address(addr)))
-
-        if hasattr(msg, "custom_args"):
-            for k, v in msg.custom_args.items():
-                personalization.add_custom_arg(CustomArg(k, v))
-
-        if self._is_transaction_template(msg):
-            if msg.subject:
-                logger.warning(
-                    "Message subject is ignored in transactional template, "
-                    "please add it as template variable (e.g. {{ subject }}"
-                )
-                # See https://github.com/sendgrid/sendgrid-nodejs/issues/843
-        else:
-            personalization.subject = msg.subject
-
+        personalization_headers = []
         for k, v in msg.extra_headers.items():
             if k.lower() == "reply-to":
                 mail.reply_to = Email(v)
             else:
-                personalization.add_header(Header(k, v))
+                personalization_headers.append(Header(k, v))
 
         if hasattr(msg, "ip_pool_name"):
             if not isinstance(msg.ip_pool_name, str):
@@ -302,17 +379,6 @@ class SendgridBackend(BaseEmailBackend):
                 ip_pool_name = IpPoolName(msg.ip_pool_name)
             mail.ip_pool_name = ip_pool_name
 
-        # write through the send_at attribute
-        if hasattr(msg, "send_at"):
-            if not isinstance(msg.send_at, int):
-                raise ValueError(
-                    "send_at must be an integer, got: {}; "
-                    "see https://sendgrid.com/docs/API_Reference/SMTP_API/scheduling_parameters.html#-Send-At".format(
-                        type(msg.send_at)
-                    )
-                )
-            personalization.send_at = msg.send_at
-
         if hasattr(msg, "reply_to") and msg.reply_to:
             if mail.reply_to:
                 # If this code path is triggered, the reply_to on the sg mail was set in a header above
@@ -348,15 +414,6 @@ class SendgridBackend(BaseEmailBackend):
         if hasattr(msg, "template_id"):
             # Template mails should not have subject and content attributes
             mail.template_id = msg.template_id
-            if hasattr(msg, "substitutions"):
-                for k, v in msg.substitutions.items():
-                    personalization.add_substitution(Substitution(k, v))
-            if hasattr(msg, "dynamic_template_data"):
-                if SENDGRID_5:
-                    logger.warning(
-                        "dynamic_template_data not available in sendgrid version < 6"
-                    )
-                personalization.dynamic_template_data = msg.dynamic_template_data
 
         if not self._is_transaction_template(msg):
             # In sendgrid v6 we should not specify subject and content between request parameter
@@ -373,7 +430,37 @@ class SendgridBackend(BaseEmailBackend):
             else:
                 mail.add_content(Content("text/plain", msg.body))
 
-        mail.add_personalization(personalization)
+        if hasattr(msg, "personalizations"):
+            for personalization in msg.personalizations:
+                if type(personalization) == Dict:
+                    personalization = dict_to_personalization(personalization)
+
+                assert type(personalization) == Personalization
+
+                mail.add_personalization(
+                    self._build_sg_personalization(
+                        msg,
+                        personalization_headers,
+                        existing_personalizations=personalization,
+                    )
+                )
+        elif getattr(msg, "make_private", False):
+            for to in msg.to:
+                mail.add_personalization(
+                    self._build_sg_personalization(
+                        msg,
+                        personalization_headers,
+                        to=[to],
+                    )
+                )
+        else:
+            mail.add_personalization(
+                self._build_sg_personalization(
+                    msg,
+                    personalization_headers,
+                    to=msg.to,
+                )
+            )
 
         if hasattr(msg, "categories"):
             for cat in msg.categories:
